/*
 * Mach Operating System
 * Copyright (c) 1991 Carnegie Mellon University
 * Copyright (c) 1991 OMRON Corporation
 * All Rights Reserved.
 *
 * Permission to use, copy, modify and distribute this software and its
 * documentation is hereby granted, provided that both the copyright
 * notice and this permission notice appear in all copies of the
 * software, derivative works or modified versions, and any portions
 * thereof, and that both notices appear in supporting documentation.
 *
 * CARNEGIE MELLON AND OMRON ALLOW FREE USE OF THIS SOFTWARE IN ITS "AS IS"
 * CONDITION.  CARNEGIE MELLON AND OMRON DISCLAIM ANY LIABILITY OF ANY KIND
 * FOR ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 *
 * Carnegie Mellon requests users of this software to return to
 *
 *  Software Distribution Coordinator  or  Software.Distribution@CS.CMU.EDU
 *  School of Computer Science
 *  Carnegie Mellon University
 *  Pittsburgh PA 15213-3890
 *
 * any improvements or extensions that they make and grant Carnegie the
 * rights to redistribute these changes.
 */

/*
 * Copyright (c) 1982,1986 Regents of the University of California.
 * All rights reserved.  The Berkeley software License Agreement
 * specifies the terms and conditions for redistribution.
 *
 *	@(#)autoconf.c	7.1 (Berkeley) 6/6/86
 */

/*
 * Setup the system to run on the current machine.
 *
 * Configure() is called at boot time and initializes the uba and mba
 * device tables and the memory controller monitoring.  Available
 * devices are determined (from possibilities mentioned in ioconf.c),
 * and the drivers are initialized.
 */

#include <cpus.h>

#include <sys/reboot.h>
#include <sys/types.h>
#include <mach/std_types.h>
#include <mach/machine.h>
#include <machine/custom.h>
#include <machine/cia.h>
#ifdef A3000
#include <machine/a3000.h>
#endif
#ifdef A2091
#include <machine/a2091.h>
#endif
#ifdef GVP11
#include <machine/gvp11.h>
#endif

#include <mach/machine.h>
#include <vm/vm_kern.h>
#include <mach/vm_param.h>

#include <chips/busses.h>

extern struct bus_driver sbic_driver;
/* extern struct bus_device se_driver; */

#if 0
extern void se_intr();
#endif
extern void sbic_intr();
#ifdef A3000
extern void a3000_init();
#endif
#ifdef A2091
extern void a2091_init();
#endif
#ifdef GVP11
extern void gvp11_init ();
#endif

extern int scsi_debug;

struct bus_ctlr bus_master_init[] = {
/* driver, name, unit, intr, addr+am, phys, adaptor, alive, flags */
{ &sbic_driver, "sbic", 0, sbic_intr, 0x0, 0, 0, '?', 0, 0,} ,
  0
};

struct bus_device bus_device_init[] = {
{ &sbic_driver, "rz", 0, 0, 0x0, 0, 0, '?', 0, 0, 0, 0, },
{ &sbic_driver, "rz", 1, 0, 0x0, 0, 0, '?', 0, 0, 1, 0, },
{ &sbic_driver, "rz", 2, 0, 0x0, 0, 0, '?', 0, 0, 2, 0, },
{ &sbic_driver, "rz", 3, 0, 0x0, 0, 0, '?', 0, 0, 3, 0, },
{ &sbic_driver, "rz", 4, 0, 0x0, 0, 0, '?', 0, 0, 4, 0, },
{ &sbic_driver, "rz", 5, 0, 0x0, 0, 0, '?', 0, 0, 5, 0, },
{ &sbic_driver, "rz", 6, 0, 0x0, 0, 0, '?', 0, 0, 6, 0, },
{ &sbic_driver, "rz", 7, 0, 0x0, 0, 0, '?', 0, 0, 7, 0, },
{ &sbic_driver, "tz", 0, 0, 0x0, 0, 0, '?', 0, 0, 0, 0, },
{ &sbic_driver, "tz", 1, 0, 0x0, 0, 0, '?', 0, 0, 1, 0, },
{ &sbic_driver, "tz", 2, 0, 0x0, 0, 0, '?', 0, 0, 2, 0, },
{ &sbic_driver, "tz", 3, 0, 0x0, 0, 0, '?', 0, 0, 3, 0, },
{ &sbic_driver, "tz", 4, 0, 0x0, 0, 0, '?', 0, 0, 4, 0, },
{ &sbic_driver, "tz", 5, 0, 0x0, 0, 0, '?', 0, 0, 5, 0, },
{ &sbic_driver, "tz", 6, 0, 0x0, 0, 0, '?', 0, 0, 6, 0, },
{ &sbic_driver, "tz", 7, 0, 0x0, 0, 0, '?', 0, 0, 7, 0, },
#if 0
/* lance */
{ &se_driver,  "am", 0, se_intr, 0x0, 0, 0, '?', 0, -1, -1, 0, },
#endif
  0
};

/*
 * The following several variables are related to
 * the configuration process, and are used in initializing
 * the machine.
 */
#ifdef unused
u_short	*caddr;		/* address being probed */
int	devtype;
int	nexnum;		/* current nexus number */
int	nsbi;		/* current sbi number */
#endif
int	cold = 1;	/* if 1, still working on cold-start */
int	dkn;		/* number of iostat dk numbers assigned so far */

extern int loadpt;

extern char *bdevname[];	/* see ../conf/param.c */
extern int nbdevnames;		/* see ../conf/param.c */
extern char *clu_name_list[];	/* see ../conf/param.c */




/*
 * Determine mass storage and memory configuration for a machine.
 * Get cpu type, and then switch out to machine specific procedures
 * which will probe adaptors to see what is out there.
 */
void machine_init()
{
  int i;
  extern int	cpu_sets[];
  extern int	master_cpu;
  extern int	hz;

  scsi_debug = 1;

printf("machine_init\n");

  master_cpu = 0;
  machine_slot[0].is_cpu = TRUE;
  machine_slot[0].running = TRUE;
  machine_slot[0].clock_freq = hz;
 /* We don't really know, do we */
  machine_slot[0].cpu_type = CPU_TYPE_MC68030;
#ifdef A3000
  machine_slot[0].cpu_subtype = CPU_SUBTYPE_AMIGA_3000;

  a3000_init();
#else
 /* We don't really know, do we */
  machine_slot[0].cpu_subtype = CPU_SUBTYPE_AMIGA_2000;

#ifdef A2091
  a2091_init();
#endif
#ifdef GVP11
  gvp11_init ();
#endif
#endif

  custom.intena = 0x7fff;

  /* clear any possibly pending interrupts */
  custom.intreq = 0x7fff;
  /* really! (-> bloody keyboard!) */
  *(volatile char *)&ciaa.icr;
  *(volatile char *)&ciab.icr;
 
  /* enable:
     SOFTINT:   software interrupt request
     VERTB:     start of Vertical Blank
     RBF:       serial port Receive Buffer Full
     TBE:       serial port Transmit Buffer Empty
     PORTS:     CIA A interrupt (keyboard and parallel port)
     EXTER:     CIA B interrupt (hardclock) */
  custom.intena = (INTF_SETCLR | INTF_INTEN | INTF_SOFTINT /*| INTF_VERTB*/
		   | INTF_RBF | INTF_TBE | INTF_PORTS | INTF_EXTER);
 
printf ("about to enable interrupts\n");
 
  spl0();

printf ("spl0, configure_bus_master\n");

  /* Probe the sbic based scsi busses (with ints enabled!) */
#ifdef A3000
  configure_bus_master("sbic", (caddr_t)A3000_SCSI_ADDR, 
	               (caddr_t)PHYS_A3000_SCSI_ADDR, 0, "A3000 on board SCSI");
#endif
#ifdef A2091
  configure_bus_master("sbic", (caddr_t)A2091_SCSI_ADDR, 
	               (caddr_t)PHYS_A2091_SCSI_ADDR, 0, "A2091/A590 SCSI Controller");
#endif
#ifdef GVP11
  configure_bus_master("sbic", (caddr_t)GVP11_SCSI_ADDR, 
	               (caddr_t)PHYS_GVP11_SCSI_ADDR, 0, "GVP series II (prod. 11) SCSI");
#endif
}

