/*
 * Mach Operating System
 * Copyright (c) 1991,1990,1989 Carnegie Mellon University
 * All Rights Reserved.
 *
 * Permission to use, copy, modify and distribute this software and its
 * documentation is hereby granted, provided that both the copyright
 * notice and this permission notice appear in all copies of the
 * software, derivative works or modified versions, and any portions
 * thereof, and that both notices appear in supporting documentation.
 *
 * CARNEGIE MELLON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS"
 * CONDITION.  CARNEGIE MELLON DISCLAIMS ANY LIABILITY OF ANY KIND FOR
 * ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 *
 * Carnegie Mellon requests users of this software to return to
 *
 *  Software Distribution Coordinator  or  Software.Distribution@CS.CMU.EDU
 *  School of Computer Science
 *  Carnegie Mellon University
 *  Pittsburgh PA 15213-3890
 *
 * any improvements or extensions that they make and grant Carnegie Mellon
 * the rights to redistribute these changes.
 */
/*
 * Amiga keyboard/display driver
 *
 * Contributors:
 *      Bryan Ford
 */

#include <mach/std_types.h>
#include <mach/machine.h>
#include <machine/pmap.h>
#include <machine/custom.h>
#include <machine/cia.h>
#include <machine/kd.h>
#include <machine/machparam.h>
#include <device/tty.h>
#include <device/io_req.h>
#include <device/device_types.h>
#include <device/buf.h>
#include <device/cirbuf.h>

/* Display sizes in characters */
#define NTSC_WIDTH      80
#define NTSC_HEIGHT     27

#define PAL_WIDTH       80
#define PAL_HEIGHT      34


extern unsigned char kernel_font[];
#define FONT_WIDTH      8       /* This width is "assumed" in several places below.  */
#define FONT_HEIGHT     15

unsigned char *screen_mem;
int screen_width, screen_height;

int px,py;

struct tty kd_tty;
extern struct cirbuf *cons_buf;

/*
 * kd_state shows the state of the modifier keys (ctrl, caps lock,
 * etc.)  It should normally be changed by calling set_kd_state(), so
 * that the keyboard status LEDs are updated correctly.
 */
int   kd_state  = KS_NORMAL;
int kb_mode   = KB_ASCII; /* event/ascii */


/* This array maps scancodes to Ascii characters (or character sequences).
   Each row corresponds to one key.  There are NUMOUTPUT bytes per key
   state.  The states are ordered: Normal, SHIFT, CTRL, ALT, SHIFT/ALT.  */

/* This keymap probably needs a great deal of debugging.  */
unsigned char   key_map[NUMKEYS][WIDTH_KMAP] = {

/* 0 */
{K_GRAV,NC,NC, K_TILDE,NC,NC, K_RS,NC,NC, 0x1b,0x4e,0x60, 0x1b,0x4e,0x7e},
{K_ONE,NC,NC, K_BANG,NC,NC, K_ONE,NC,NC, 0x1b,0x4e,0x31, 0x1b,0x4e,0x21},
{K_TWO,NC,NC, K_ATSN,NC,NC, K_NUL,NC,NC, 0x1b,0x4e,0x32, 0x1b,0x4e,0x40},
{K_THREE,NC,NC, K_POUND,NC,NC, K_THREE,NC,NC, 0x1b,0x4e,0x33, 0x1b,0x4e,0x23},
{K_FOUR,NC,NC, K_DOLLAR,NC,NC, K_FOUR,NC,NC, 0x1b,0x4e,0x34, 0x1b,0x4e,0x24},
{K_FIVE,NC,NC, K_PERC,NC,NC, K_FIVE,NC,NC, 0x1b,0x4e,0x35, 0x1b,0x4e,0x25},
{K_SIX,NC,NC, K_CARET,NC,NC, K_RS,NC,NC, 0x1b,0x4e,0x36, 0x1b,0x4e,0x5e},
{K_SEVEN,NC,NC, K_AMPER,NC,NC, K_SEVEN,NC,NC, 0x1b,0x4e,0x37, 0x1b,0x4e,0x26},
{K_EIGHT,NC,NC, K_ASTER,NC,NC, K_EIGHT,NC,NC, 0x1b,0x4e,0x38, 0x1b,0x4e,0x2a},
{K_NINE,NC,NC, K_LPAREN,NC,NC, K_NINE,NC,NC, 0x1b,0x4e,0x39,0x1b,0x4e,0x28},
{K_ZERO,NC,NC, K_RPAREN,NC,NC, K_ZERO,NC,NC, 0x1b,0x4e,0x30,0x1b,0x4e,0x29},
{K_MINUS,NC,NC, K_UNDSC,NC,NC, K_US,NC,NC, 0x1b,0x4e,0x2d, 0x1b,0x4e,0x5f},
{K_EQL,NC,NC, K_PLUS,NC,NC, K_EQL,NC,NC, 0x1b,0x4e,0x3d, 0x1b,0x4e,0x2b},
{K_BSLSH,NC,NC, K_PIPE,NC,NC, K_FS,NC,NC, 0x1b,0x4e,0x5c, 0x1b,0x4e,0x7c},
{NC,NC,NC,NC,NC,NC,NC,NC,NC,NC,NC,NC,NC,NC,NC},
{K_ZERO,NC,NC, K_ZERO,NC,NC, K_ZERO,NC,NC, 0x1b,0x4e,0x30,0x1b,0x4e,0x29},

/* 1 */
{K_q,NC,NC, K_Q,NC,NC, K_DC1,NC,NC, 0x1b,0x4e,0x71, 0x1b,0x4e,0x51},
{K_w,NC,NC, K_W,NC,NC, K_ETB,NC,NC, 0x1b,0x4e,0x77, 0x1b,0x4e,0x57},
{K_e,NC,NC, K_E,NC,NC, K_ENQ,NC,NC, 0x1b,0x4e,0x65, 0x1b,0x4e,0x45},
{K_r,NC,NC, K_R,NC,NC, K_DC2,NC,NC, 0x1b,0x4e,0x72, 0x1b,0x4e,0x52},
{K_t,NC,NC, K_T,NC,NC, K_DC4,NC,NC, 0x1b,0x4e,0x74, 0x1b,0x4e,0x54},
{K_y,NC,NC, K_Y,NC,NC, K_EM,NC,NC, 0x1b,0x4e,0x79, 0x1b,0x4e,0x59},
{K_u,NC,NC, K_U,NC,NC, K_NAK,NC,NC, 0x1b,0x4e,0x75, 0x1b,0x4e,0x55},
{K_i,NC,NC, K_I,NC,NC, K_HT,NC,NC, 0x1b,0x4e,0x69, 0x1b,0x4e,0x49},
{K_o,NC,NC, K_O,NC,NC, K_SI,NC,NC, 0x1b,0x4e,0x6f, 0x1b,0x4e,0x4f},
{K_p,NC,NC, K_P,NC,NC, K_DLE,NC,NC, 0x1b,0x4e,0x70, 0x1b,0x4e,0x50},
{K_LBRKT,NC,NC, K_LBRACE,NC,NC, K_ESC,NC,NC, 0x1b,0x4e,0x5b, 0x1b,0x4e,0x7b},
{K_RBRKT,NC,NC, K_RBRACE,NC,NC, K_GS,NC,NC, 0x1b,0x4e,0x5d, 0x1b,0x4e,0x7d},
{NC,NC,NC,NC,NC,NC,NC,NC,NC,NC,NC,NC,NC,NC,NC},
{K_ONE,NC,NC, K_ONE,NC,NC, K_ONE,NC,NC, 0x1b,0x4e,0x31, 0x1b,0x4e,0x21},
{K_TWO,NC,NC, K_TWO,NC,NC, K_NUL,NC,NC, 0x1b,0x4e,0x32, 0x1b,0x4e,0x40},
{K_THREE,NC,NC, K_THREE,NC,NC, K_THREE,NC,NC, 0x1b,0x4e,0x33, 0x1b,0x4e,0x23},

/* 2 */
{K_a,NC,NC, K_A,NC,NC, K_SOH,NC,NC, 0x1b,0x4e,0x61, 0x1b,0x4e,0x41},
{K_s,NC,NC, K_S,NC,NC, K_DC3,NC,NC, 0x1b,0x4e,0x73, 0x1b,0x4e,0x53},
{K_d,NC,NC, K_D,NC,NC, K_EOT,NC,NC, 0x1b,0x4e,0x65, 0x1b,0x4e,0x45},
{K_f,NC,NC, K_F,NC,NC, K_ACK,NC,NC, 0x1b,0x4e,0x66, 0x1b,0x4e,0x46},
{K_g,NC,NC, K_G,NC,NC, K_BEL,NC,NC, 0x1b,0x4e,0x67, 0x1b,0x4e,0x47},
{K_h,NC,NC, K_H,NC,NC, K_BS,NC,NC, 0x1b,0x4e,0x68, 0x1b,0x4e,0x48},
{K_j,NC,NC, K_J,NC,NC, K_LF,NC,NC, 0x1b,0x4e,0x6a, 0x1b,0x4e,0x4a},
{K_k,NC,NC, K_K,NC,NC, K_VT,NC,NC, 0x1b,0x4e,0x6b, 0x1b,0x4e,0x4b},
{K_l,NC,NC, K_L,NC,NC, K_FF,NC,NC, 0x1b,0x4e,0x6c, 0x1b,0x4e,0x4c},
{K_SEMI,NC,NC, K_COLON,NC,NC, K_SEMI,NC,NC, 0x1b,0x4e,0x3b, 0x1b,0x4e,0x3a},
{K_SQUOTE,NC,NC,K_DQUOTE,NC,NC,K_SQUOTE,NC,NC,0x1b,0x4e,0x27,0x1b,0x4e,0x22},
{NC,NC,NC,NC,NC,NC,NC,NC,NC,NC,NC,NC,NC,NC,NC},
{NC,NC,NC,NC,NC,NC,NC,NC,NC,NC,NC,NC,NC,NC,NC},
{K_FOUR,NC,NC, K_FOUR,NC,NC, K_FOUR,NC,NC, 0x1b,0x4e,0x34, 0x1b,0x4e,0x24},
{K_FIVE,NC,NC, K_FIVE,NC,NC, K_FIVE,NC,NC, 0x1b,0x4e,0x35, 0x1b,0x4e,0x25},
{K_SIX,NC,NC, K_SIX,NC,NC, K_RS,NC,NC, 0x1b,0x4e,0x36, 0x1b,0x4e,0x5e},

/* 3 */
{NC,NC,NC,NC,NC,NC,NC,NC,NC,NC,NC,NC,NC,NC,NC},
{K_z,NC,NC, K_Z,NC,NC, K_SUB,NC,NC, 0x1b,0x4e,0x7a, 0x1b,0x4e,0x5a},
{K_x,NC,NC, K_X,NC,NC, K_CAN,NC,NC, 0x1b,0x4e,0x78, 0x1b,0x4e,0x58},
{K_c,NC,NC, K_C,NC,NC, K_ETX,NC,NC, 0x1b,0x4e,0x63, 0x1b,0x4e,0x43},
{K_v,NC,NC, K_V,NC,NC, K_SYN,NC,NC, 0x1b,0x4e,0x76, 0x1b,0x4e,0x56},
{K_b,NC,NC, K_B,NC,NC, K_STX,NC,NC, 0x1b,0x4e,0x62, 0x1b,0x4e,0x42},
{K_n,NC,NC, K_N,NC,NC, K_SO,NC,NC, 0x1b,0x4e,0x6e, 0x1b,0x4e,0x4e},
{K_m,NC,NC, K_M,NC,NC, K_CR,NC,NC, 0x1b,0x4e,0x6d, 0x1b,0x4e,0x4d},
{K_COMMA,NC,NC, K_LTHN,NC,NC, K_COMMA,NC,NC, 0x1b,0x4e,0x2c, 0x1b,0x4e,0x3c},
{K_PERIOD,NC,NC, K_GTHN,NC,NC, K_PERIOD,NC,NC,0x1b,0x4e,0x2e,0x1b,0x4e,0x3e},
{K_SLASH,NC,NC, K_QUES,NC,NC, K_SLASH,NC,NC, 0x1b,0x4e,0x2f, 0x1b,0x4e,0x3f},
{NC,NC,NC,NC,NC,NC,NC,NC,NC,NC,NC,NC,NC,NC,NC},
{K_PERIOD,NC,NC, K_PERIOD,NC,NC, K_PERIOD,NC,NC,0x1b,0x4e,0x2e,0x1b,0x4e,0x3e},
{K_SEVEN,NC,NC, K_SEVEN,NC,NC, K_SEVEN,NC,NC, 0x1b,0x4e,0x37, 0x1b,0x4e,0x26},
{K_EIGHT,NC,NC, K_EIGHT,NC,NC, K_EIGHT,NC,NC, 0x1b,0x4e,0x38, 0x1b,0x4e,0x2a},
{K_NINE,NC,NC, K_NINE,NC,NC, K_NINE,NC,NC, 0x1b,0x4e,0x39,0x1b,0x4e,0x28},

/* 4 */
{K_SPACE,NC,NC, K_SPACE,NC,NC, K_NUL,NC,NC, K_SPACE,NC,NC, K_SPACE,NC,NC},
{K_BS,NC,NC, K_BS,NC,NC, K_BS,NC,NC, K_BS,NC,NC, K_BS,NC,NC},
{NC,NC,NC,NC,NC,NC,NC,NC,NC,NC,NC,NC,NC,NC,NC},
{K_CR,NC,NC, K_CR,NC,NC, K_CR,NC,NC, K_CR,NC,NC, K_CR,NC,NC},
{K_CR,NC,NC, K_CR,NC,NC, K_CR,NC,NC, K_CR,NC,NC, K_CR,NC,NC},
{K_ESC,NC,NC, K_ESC,NC,NC, K_ESC,NC,NC, K_ESC,NC,NC, K_ESC,NC,NC},
{K_DEL,NC,NC, K_PERIOD,NC,NC, K_DEL,NC,NC, K_DEL,NC,NC, 0x1b,0x4e,0x2e},
{NC,NC,NC,NC,NC,NC,NC,NC,NC,NC,NC,NC,NC,NC,NC},
{NC,NC,NC,NC,NC,NC,NC,NC,NC,NC,NC,NC,NC,NC,NC},
{NC,NC,NC,NC,NC,NC,NC,NC,NC,NC,NC,NC,NC,NC,NC},
{K_MINUS,NC,NC, K_MINUS,NC,NC, K_MINUS,NC,NC, 0x1b,0x4e,0x2d, 0x1b,0x4e,0x5f},
{NC,NC,NC,NC,NC,NC,NC,NC,NC,NC,NC,NC,NC,NC,NC},
{K_UA, K_UA, K_UA, K_UA, 0x1b,0x4e,0x38},
{K_DA, K_DA, K_DA, K_DA, 0x1b,0x4e,0x32},
{K_RA, K_RA, K_RA, K_RA, 0x1b,0x4e,0x36},
{K_LA, K_LA, K_LA, K_LA, 0x1b,0x4e,0x34},

/* 5 */
{K_F1, K_F1S, K_F1, K_F1, K_F1S},
{K_F2, K_F2S, K_F2, K_F2, K_F2S},
{K_F3, K_F3S, K_F3, K_F3, K_F3S},
{K_F4, K_F4S, K_F4, K_F4, K_F4S},
{K_F5, K_F5S, K_F5, K_F5, K_F5S},
{K_F6, K_F6S, K_F6, K_F6, K_F6S},
{K_F7, K_F7S, K_F7, K_F7, K_F7S},
{K_F8, K_F8S, K_F8, K_F8, K_F8S},
{K_F9, K_F9S, K_F9, K_F9, K_F9S},
{K_F10, K_F10S, K_F10, K_F10, K_F10S},
{NC,NC,NC,NC,NC,NC,NC,NC,NC,NC,NC,NC,NC,NC,NC},
{NC,NC,NC,NC,NC,NC,NC,NC,NC,NC,NC,NC,NC,NC,NC},
{NC,NC,NC,NC,NC,NC,NC,NC,NC,NC,NC,NC,NC,NC,NC},
{NC,NC,NC,NC,NC,NC,NC,NC,NC,NC,NC,NC,NC,NC,NC},
{NC,NC,NC,NC,NC,NC,NC,NC,NC,NC,NC,NC,NC,NC,NC},
{NC,NC,NC,NC,NC,NC,NC,NC,NC,NC,NC,NC,NC,NC,NC}, /* XXX help key */

/* 6 */
{K_SCAN,K_LSHSC,NC, K_SCAN,K_LSHSC,NC, K_SCAN,K_LSHSC,NC, K_SCAN,K_LSHSC,NC, K_SCAN,K_LSHSC,NC},
{K_SCAN,K_RSHSC,NC, K_SCAN,K_RSHSC,NC, K_SCAN,K_RSHSC,NC, K_SCAN,K_RSHSC,NC, K_SCAN,K_RSHSC,NC},
{K_SCAN,K_CLKSC,NC, K_SCAN,K_CLKSC,NC, K_SCAN,K_CLKSC,NC, K_SCAN,K_CLKSC,NC, K_SCAN,K_CLKSC,NC},
{K_SCAN,K_CTLSC,NC, K_SCAN,K_CTLSC,NC, K_SCAN,K_CTLSC,NC, K_SCAN,K_CTLSC,NC, K_SCAN,K_CTLSC,NC},
{K_SCAN,K_LALTSC,NC, K_SCAN,K_LALTSC,NC, K_SCAN,K_LALTSC,NC, K_SCAN,K_LALTSC,NC, K_SCAN,K_LALTSC,NC},
{K_SCAN,K_RALTSC,NC, K_SCAN,K_RALTSC,NC, K_SCAN,K_RALTSC,NC, K_SCAN,K_RALTSC,NC, K_SCAN,K_RALTSC,NC},
{NC,NC,NC,NC,NC,NC,NC,NC,NC,NC,NC,NC,NC,NC,NC}, /* XXX amiga keys */
{NC,NC,NC,NC,NC,NC,NC,NC,NC,NC,NC,NC,NC,NC,NC}

};



/* Clear the screen */
static void
clear(void)
{
  bzero(screen_mem,screen_width*screen_height*FONT_HEIGHT);
  px = py = 0;
}


static void
scroll_up(void)
{
  ovbcopy(screen_mem+screen_width*FONT_HEIGHT,screen_mem,screen_width*(screen_height-1)*FONT_HEIGHT);
  bzero(screen_mem+screen_width*(screen_height-1)*FONT_HEIGHT,screen_width*FONT_HEIGHT);
  if(--py < 0)
    py = 0;
}


static void
xorcurs(void)
{
  unsigned char *op = screen_mem+px+py*screen_width*FONT_HEIGHT;
  int i;

  for(i = FONT_HEIGHT;i > 0;i--)
    *op = ~(*op), op += screen_width;
}



int
kd_putc(int c)
{
  u_int s = spltty ();

  xorcurs();
  switch(c)
    {
      case '\n':
        linefeed:
        if(++py >= screen_height)
          scroll_up();
        break;
      case '\b':
        if(px > 0)
          px--;
        break;
      case '\r':
        px = 0;
        break;
      case '\f':
        clear();
        break;
      case '\t':
        do
          {
            kd_putc(' ');
          }
        while(px & 7);
        break;
      default:
        if(c >= ' ')
          {
            unsigned char *ip = &kernel_font[(c-32)*FONT_HEIGHT];
            unsigned char *op = screen_mem+px+py*screen_width*FONT_HEIGHT;
            int i;

            for(i = FONT_HEIGHT;i > 0;i--)
              *op = *ip++, op += screen_width;

            if(++px >= screen_width)
              {
                newline:
                px = 0;
                goto linefeed;
              }
          }
        break;
    }
  xorcurs();
  splx(s);
  return(0);
}


/* Called by cnputc, to write to the console from the kernel */
void
kd_cnputc(int c)
{
  kd_putc(c);
}



/* Initialize one copper list.  We'll need two to make a nice interlaced display.  */

#define COPENTS 13 /* Number of instruction's we'll need in each copper list */

static void initcop(unsigned short *cop,unsigned short *othercop,int shf)
{
  long scrmem = (long)screen_mem + shf*screen_width;

  *cop++ = CUSTOM_OFS(bplpt[0]);
  *cop++ = (long)scrmem >> 16;
  *cop++ = CUSTOM_OFS(bplpt[0]) + 2;
  *cop++ = (long)scrmem;
  *cop++ = CUSTOM_OFS(bpl1mod);
  *cop++ = screen_width;
  *cop++ = CUSTOM_OFS(bplcon0);
  *cop++ = 0x9204;
  *cop++ = CUSTOM_OFS(diwstrt);
  *cop++ = 0x2c81;
  *cop++ = CUSTOM_OFS(diwstop);
  *cop++ = 0x2cc1 + (((screen_height*FONT_HEIGHT+1-shf)>>1)<<8);
  *cop++ = CUSTOM_OFS(ddfstrt);
  *cop++ = 0x003c;
  *cop++ = CUSTOM_OFS(ddfstop);
  *cop++ = 0x00d4;
  *cop++ = CUSTOM_OFS(cop1lc);
  *cop++ = (long)othercop >> 16;
  *cop++ = CUSTOM_OFS(cop1lc) + 2;
  *cop++ = (long)othercop;
  *cop++ = CUSTOM_OFS(color[0]);
  *cop++ = 0x123;
  *cop++ = CUSTOM_OFS(color[1]);
  *cop++ = 0xccc;
  *(int*)cop = 0xfffffffe;
}



/* Initialize the screen and keyboard.  Called early from cons.c.  */
void
kdinit(void)
{
  int nvpos, ovpos;
  void *coplist;

  /* See if we have NTSC or PAL */
  ovpos = 0;
  while(1)
    {
      nvpos = *((int*)&custom.vposr) & 0x0001ff00;
      if(nvpos < ovpos)
        break;
      ovpos = nvpos;
    }

  if(ovpos < (270 << 8))
    screen_width = NTSC_WIDTH, screen_height = NTSC_HEIGHT;
  else
    screen_width = PAL_WIDTH, screen_height = PAL_HEIGHT;


  screen_mem = (void*)chipmem_steal(screen_width*screen_height*FONT_HEIGHT);

  clear();
  xorcurs();

  coplist = (void*)chipmem_steal(2*COPENTS*4);
  initcop(coplist,coplist+COPENTS*4,0);
  initcop(coplist+COPENTS*4,coplist,1);

  /* Make sure no ex-sprites are streaking down the screen */
  {
    int i;
    for(i = 0;i < 8;i++)
      {
        custom.spr[i].data = 0;
        custom.spr[i].datb = 0;
      }
  }

  /* Start the new display */
  custom.dmacon = 0x8380;
  custom.cop1lc = coplist;
  custom.copjmp1 = 0;
  while((signed short)custom.vposr >= 0); /* Make sure we're in a long frame */
  custom.cop1lc = coplist;
  custom.copjmp1 = 0;
}



/*
 * kdstart:
 *
 *  This function does the general processing of characters and other
 *  operations for the device driver.  The device independent portion of
 *  the tty driver calls this routine (it's setup in kdinit) with a
 *  given command.  That command is then processed, and control is passed
 *  back to the kernel.
 *
 * input:   tty pointer 'tp', and command to execute 'cmd'
 *
 * output:  command is executed
 *
 * Entered and left at spltty.  Drops priority to spl0 to display character.
 * ASSUMES that it is never called from interrupt-driven code.
 */
kdstart(struct tty *tp)
{
    int o_pri;
    int ch;
    unsigned char   c;

    if (tp->t_state & TS_TTSTOP)
        return;
    for ( ; ; ) {
        tp->t_state &= ~TS_BUSY;
        if (tp->t_state & TS_TTSTOP)
            break;
        if ((tp->t_outq.c_cc <= 0) || (ch = getc(&tp->t_outq)) == -1)
            break;
        c = ch;
        /*
         * Drop priority for long screen updates. ttstart() calls us at
         * spltty.
         */
        o_pri = splsoftclock();     /* block timeout */
        /* XXX - escape codes */
        kd_putc(c);
        splx(o_pri);
    }
    if (tp->t_outq.c_cc <= TTLOWAT(tp)) {
        tt_write_wakeup(tp);
    }
    return;
}

kdstop(struct tty *tp, int flags)
{
    /*
     * do nothing - all characters are output by one call to
     * kdstart.
     */
}

/*
 * kdopen:
 *
 *  This opens the console driver and sets up the tty and other
 *  rudimentary stuff including calling the line discipline for
 *  setting up the device independent stuff for a tty driver.
 *
 * input:   device number 'dev', and flag
 *
 * output:  device is opened and setup
 *
 */
kdopen(dev_t dev, int flag, io_req_t ior)
{
  struct tty *tp;
  int o_pri;

  tp = &kd_tty;
  o_pri = spltty();
  simple_lock(&tp->t_lock);
  if (!(tp->t_state & (TS_ISOPEN|TS_WOPEN))) {
      ttychars(tp);
      /*
       *  Special support for boot-time rc scripts, which don't
       *  stty the console.
       */
      tp->t_oproc = kdstart;
      tp->t_stop = kdstop;
      tp->t_ospeed = tp->t_ispeed = B9600;
      tp->t_flags = ODDP|EVENP|ECHO|CRMOD;
  }
  tp->t_state |= TS_CARR_ON;
  simple_unlock(&tp->t_lock);
  splx(o_pri);
  return (char_open(dev, tp, flag, ior));
}


/*
 * kdclose:
 *
 *  This function merely executes the device independent code for
 *  closing the line discipline.
 *
 * input:   device number 'dev', and flag
 *
 * output:  device is closed
 *
 */
kdclose(int dev, int flag)
{
    struct  tty *tp;

    tp = &kd_tty;
    {
        int s = spltty();
        simple_lock(&tp->t_lock);
        ttyclose(tp);
        simple_unlock(&tp->t_lock);
        splx(s);
    }
    return;

}


/*
 * kdread:
 *
 *  This function executes the device independent code to read from
 *  the tty.
 *
 * input:   device number 'dev'
 *
 * output:  characters are read from tty clists
 *
 */
kdread(int dev, struct uio *uio)
{
    struct  tty *tp;

    tp = &kd_tty;
    tp->t_state |= TS_CARR_ON;
    return((*linesw[kd_tty.t_line].l_read)(tp, uio));
}


/*
 * kdwrite:
 *
 *  This function does the device independent write action for this
 *  console (tty) driver.
 *
 * input:   device number 'dev'
 *
 * output:  characters are written to tty clists
 *
 */
kdwrite(int dev, struct uio *uio)
{
    return((*linesw[kd_tty.t_line].l_write)(&kd_tty, uio));
}


int
kdmmap(dev_t dev, off_t off, int prot)
{
  if ((u_int) off >= (128*1024))
    return(-1);

  /* Get page frame number for the page to be mapped. */
  return(pmap_phys_to_frame(screen_mem+off));
}


int
kdportdeath(dev, port)
    dev_t   dev;
    mach_port_t port;
{
  return (tty_portdeath(&kd_tty, port));
}


io_return_t
kdgetstat(dev_t dev, int flavor, int *data, unsigned int *count)
{
  io_return_t result;

  switch (flavor)
    {
      case KDGSTATE:
        if (*count < 1)
            return (D_INVALID_OPERATION);
        *data = kd_state;
        *count = 1;
        result = D_SUCCESS;
        break;

      case KDGKBENT:
        result = kdgetkbent((struct kbentry *)data);
        *count = sizeof(struct kbentry)/sizeof(int);
        break;

      default:
        result = tty_get_status(&kd_tty, flavor, data, count);
        break;
    }
  return (result);
}


io_return_t
kdsetstat(dev_t dev, int flavor, int *data, unsigned int count)
{
  io_return_t result;

  switch (flavor)
    {
      case KDSKBENT:
        if (count < sizeof(struct kbentry)/sizeof(int)) {
            return (D_INVALID_OPERATION);
        }
        result = kdsetkbent((struct kbentry *)data, 0);
        break;

      default:
        result = tty_set_status(&kd_tty, flavor, data, count);
    }
  return (result);
}



/*
 * kdgetkbent:
 *
 *  Get entry from key mapping table.  Returns error code, if any.
 */
int
kdgetkbent(struct kbentry *kbent)
{
  u_char *cp;
  int o_pri = SPLKD();        /* probably superfluous */

  cp = &key_map[kbent->kb_index][CHARIDX(kbent->kb_state)];
  kbent->kb_value[0] = *cp++;
  kbent->kb_value[1] = *cp++;
  kbent->kb_value[2] = *cp;
  (void)splx(o_pri);
  return(0);
}


/*
 * kdsetkbent:
 *
 *  Set entry in key mapping table.  Return error code, if any.
 */
int
kdsetkbent(struct kbentry *kbent, int flags)
{
  u_char *cp;
  int o_pri;

  o_pri = SPLKD();
  cp = &key_map[kbent->kb_index][CHARIDX(kbent->kb_state)];
  *cp++ = kbent->kb_value[0];
  *cp++ = kbent->kb_value[1];
  *cp = kbent->kb_value[2];
  (void)splx(o_pri);
  return(0);
}


/*
 * kdstate2idx:
 *
 *  Return the value for the 2nd index into key_map that
 *  corresponds to the given state.
 */

static int
kdstate2idx(int state)
{
  int state_idx = NORM_STATE;

  if(state != KS_NORMAL)
    {
      if(state & KS_ALTED)
        {
          if(state & (KS_SHIFTED | KS_CLKED))
            state_idx = SHIFT_ALT;
          else
            state_idx = ALT_STATE;
        }
      else if(state & (KS_SHIFTED | KS_CLKED))
        state_idx = SHIFT_STATE;
      else if(state & KS_CTLED)
        state_idx = CTRL_STATE;
    }
  return (CHARIDX(state_idx));
}

/*
 * do_modifier:
 *
 *  Change keyboard state according to which modifier key and
 *  whether it went down or up.
 *
 * input:   the current state, the key, and the key's direction.
 *      The key can be any key, not just a modifier key.
 *
 * output:  the new state
 */
static void
do_modifier(Scancode c, boolean_t up)
{
  switch (c)
    {
      case (K_LALTSC):
      case (K_RALTSC):
        if (up)
            kd_state &= ~KS_ALTED;
        else
            kd_state |= KS_ALTED;
        break;
      case (K_CTLSC):
        if (up)
            kd_state &= ~KS_CTLED;
        else
            kd_state |= KS_CTLED;
        break;
      case (K_CLKSC):
        if (up)
            kd_state &= ~KS_CLKED;
        else
            kd_state |= KS_CLKED;
        break;
      case (K_LSHSC):
      case (K_RSHSC):
        if (up)
            kd_state &= ~KS_SHIFTED;
        else
            kd_state |= KS_SHIFTED;
        break;
    }
}


/*
 * kdintr:
 *
 *  This function is the interrupt code for the driver.  Since this is
 *  a special tty (console), interrupts are only for input, so we read in
 *  the character.  If in ascii mode, we then do the mapping translation
 *  from the keyboard switch table and place the characters on the tty's
 *  input switch table.  If in event mode, we create and queue a kd_event.
 *
 * output:  character or sequence is placed on appropriate queue
 *
 */
void
kdintr(void)
{
  struct tty  *tp;
  unsigned char c;
  unsigned char incode, scancode;
  int o_pri;
  int char_idx;
  boolean_t up;

  tp = &kd_tty;
  o_pri = SPLKD();
  incode = ciaa.sdr;
  ciaa.cra |= 1<<6;
  ciaa.sdr = 0xff;
  scancode = ~incode;
  up = scancode & 1;
  scancode >>= 1;

  if (scancode < NUMKEYS)
    {
      /* Lookup in map, then process. */
      char_idx = kdstate2idx(kd_state);
      c = key_map[scancode][char_idx];
      if (c == K_SCAN)
        {
          c = key_map[scancode][++char_idx];
          do_modifier(c, up);
        }
      else if (!up)
        {
          /* regular key-down */
          int max;  /* max index for char sequence */

          max = char_idx + NUMOUTPUT;
          char_idx++;

          /*
           * here's where we actually put the char (or
           * char sequence, for function keys) onto the
           * input queue.
           */
          for ( ; (c != K_DONE) && (char_idx <= max);
               c = key_map[scancode][char_idx++]) {
#if 0
            (*linesw[tp->t_line].l_rint)(c, tp);
#else
	    putc (c, cons_buf);
#endif
          }
        }
    }

#if 0
  {
    int i;
    for(i = 0;i < 50;i++); /* oohhhh... */
  }
#endif

  ciaa.cra &= ~(1<<6);
  ciaa.sdr = incode;

  splx(o_pri);
}

#if 0
/* Called from cons.c, to poll for characters from the keyboard.  Called at spltty.  */
int
kd_cnpollc(void)
{
  int c = -1;
  if(ciaa.icr & (1<<3))
    {
      unsigned char incode, scancode;
      int char_idx;
      boolean_t up;

      custom.intreq = INTF_PORTS;

      incode = ciaa.sdr;
      ciaa.cra |= 1<<6;
      ciaa.sdr = 0xff;
      scancode = ~incode;
      up = scancode & 1;
      scancode >>= 1;

      if (scancode < NUMKEYS)
        {
          char_idx = kdstate2idx(kd_state);
          c = key_map[scancode][char_idx];
          if (c == K_SCAN)
            do_modifier(key_map[scancode][++char_idx], up);
        }

      {
        int i;
        for(i = 0;i < 50;i++); /* oohhhh... */
      }

      ciaa.cra &= ~(1<<6);
      ciaa.sdr = incode;

      if((c == K_SCAN) || (up))
        c = -1;
    }
  return c;
}
#endif
